#pragma once

#include <cstdint>
#include <iosfwd>

#include <glad/glad.h>
// put glad before
#include <GLFW/glfw3.h>

#include "camera.h"
#include "curve.h"
#include "resourcemanager.h"
#include "vec2i.h"
#include "viewport.h"

namespace animation {

class CurveApp {
public:
  CurveApp(int argc, char const *argv[]);
  ~CurveApp();

  bool initialize();
  bool run();

private:
  bool setupWindow();
  bool setupShaders();
  bool setupCurveBuffer();
  bool setupAxisBuffer();

  void oncePerFrame();
  void render();

  void resetCamera();

  void saveCurveToFile(std::string const &filePath);
  void loadCurveFromFile(std::string const &filePath);

  void loadCurveBuffer();
  void loadSelectedBuffer();
  void loadAxisBuffer();

  void reloadModelMatrix();
  void reloadViewMatrix();
  void reloadProjectionMatrix();
  void reloadMVPMatrix();

  void reloadMVPUniform();

  void selectPointAround(math::Vec2i const &pixel);
  void moveSelectedTo(math::Vec2i const &pixel);
  void deleteSelected();
  void subdivideSelected();
  void applyTranslationToSelected(math::Vec3f const &offset);
  int findClosestProjectPointTo(math::Vec2i const &pixel,
                                std::vector<math::Vec3f> const &points,
                                float pixelThreshold);
  void recalculateSelectedAxisOrigin();
  math::Vec3f calculateScreenPoint(math::Vec2i const &pixel,
                                   math::Vec3f const &origin);

  int addPoint(math::geometry::Curve &curve, int idx);
  int removePoint(math::geometry::Curve &curve, int idx);

  // Callback functions
  void setupCallbackFunctions();
  void onSetWindowSize(int width, int height);
  void onSetFramebufferSize(int width, int height);
  void onMouseButton(int button, int action, int mods);
  void onMouseMotion(double x, double y);
  void onKeyPress(int key, int scancode, int action, int mods);

private:
  GLFWwindow *m_window = nullptr;

  openGL::rendering::Viewport m_viewport = {{1000, 800}};
  openGL::rendering::ResourceManager m_resourceManager;

  openGL::scene::Camera m_camera;

  /* Scoped enums that easily convert to underlying type */
  struct ShaderID {
    enum : uint32_t { POINT = 0, LINE, AXIS, COUNT };
  };
  struct CurveID {
    enum : uint32_t { BASE = 0, COUNT };
  };
  struct AxisID {
    enum : uint32_t { BASE = 0, COUNT };
  };

  math::geometry::Curve m_curve;

  math::Mat4f m_M = math::identity();
  math::Mat4f m_V = math::identity();
  math::Mat4f m_P = math::identity();
  math::Mat4f m_MVP = math::identity();

  enum class InputMode { CAMERA_TURNTABLE, CAMERA_ZOOM, EDIT, NONE };
  InputMode m_mode = InputMode::NONE;
  std::vector<math::Vec3f> m_selectedOrigins = {{0.f, 0.f, 0.f}};
  std::vector<int> m_selectedID;
  math::Vec3f m_translationOffset = {0.f, 0.f, 0.f};

  std::string m_saveFilePath = "./savedCurve.txt";
  std::string m_loadFilePath = "./savedCurve.txt";
};

} // namespace animation
