#include "axisbuffer.h"

#include <cassert>

namespace openGL {
namespace rendering {

AxisBuffer::~AxisBuffer() { release(); }

AxisBuffer::AxisBuffer(GLuint vaoID, GLuint vertexColorBufferID)
    : m_vaoID(vaoID), m_vertexColorBufferID(vertexColorBufferID) {
  setupVAO();
}

AxisBuffer::AxisBuffer(AxisBuffer &&other)
    : m_vaoID(other.m_vaoID),
      m_vertexColorBufferID(other.m_vertexColorBufferID),
      m_bufferSize(other.m_bufferSize) {
  other.m_vaoID = 0;
  other.m_vertexColorBufferID = 0;
  other.m_bufferSize = 0;
}

AxisBuffer &AxisBuffer::operator=(AxisBuffer &&other) {
  if (this != &other) {
    release();
    std::swap(m_vaoID, other.m_vaoID);
    std::swap(m_vertexColorBufferID, other.m_vertexColorBufferID);
    std::swap(m_bufferSize, other.m_bufferSize);
  }
  return *this;
}

void AxisBuffer::bindVAO() { glBindVertexArray(m_vaoID); }

void AxisBuffer::setupVAO() {
  bindVAO();

  // position attribute
  glEnableVertexAttribArray(0); // layout=0
  bindBuffer();
  glVertexAttribPointer(0,                       // layout=0
                        3,                       // # of components
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // normalize
                        sizeof(math::Vec3f) * 2, // stride
                        (void *)0                // array buffer offset
  );

  // color attribute
  glEnableVertexAttribArray(1);                     // layout=1
  glVertexAttribPointer(1,                          // layout=1
                        3,                          // # of components
                        GL_FLOAT,                   // type of components
                        GL_FALSE,                   // normalize
                        sizeof(math::Vec3f) * 2,    // stride
                        (void *)sizeof(math::Vec3f) // array buffer offset
  );

  // reset to default
  glBindVertexArray(0);
}

void AxisBuffer::setupAxisBuffer(std::vector<math::Vec3f> const &vertices) {

  // 3 lines XYZ per point
  std::vector<math::Vec3f> tmpAxis;
  // 3 lines, 2 verts each, 1 color each
  tmpAxis.reserve(vertices.size() * 3 * 2 * 2);

  math::Vec3f const R{1.f, 0.f, 0.f};
  math::Vec3f const G{0.f, 1.f, 0.f};
  math::Vec3f const B{0.f, 0.f, 1.f};
  float const scale = 0.25f;

  for (auto &v : vertices) {
    tmpAxis.emplace_back(v);
    tmpAxis.emplace_back(R);
    tmpAxis.emplace_back(v + scale * R);
    tmpAxis.emplace_back(R);

    tmpAxis.emplace_back(v);
    tmpAxis.emplace_back(G);
    tmpAxis.emplace_back(v + scale * G);
    tmpAxis.emplace_back(G);

    tmpAxis.emplace_back(v);
    tmpAxis.emplace_back(B);
    tmpAxis.emplace_back(v + scale * B);
    tmpAxis.emplace_back(B);
  }

  bindBuffer();
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * tmpAxis.size(),
               tmpAxis.data(), GL_DYNAMIC_DRAW);

  m_bufferSize = vertices.size() * 3 * 2;
}

size_t AxisBuffer::size() const { return m_bufferSize; }

void AxisBuffer::release() {
  glDeleteVertexArrays(1, &m_vaoID);
  glDeleteBuffers(1, &m_vertexColorBufferID);

  m_vaoID = 0;
  m_vertexColorBufferID = 0;
}

void AxisBuffer::bindBuffer() {
  glBindBuffer(GL_ARRAY_BUFFER, m_vertexColorBufferID);
}

std::optional<AxisBuffer> makeAxisBuffer() {
  GLuint vaoID = 0;
  GLuint vertexColorBufferID = 0;
  glGenVertexArrays(1, &vaoID);
  glGenBuffers(1, &vertexColorBufferID);

  if (vaoID == 0 || vertexColorBufferID == 0)
    return std::nullopt;

  return AxisBuffer(vaoID, vertexColorBufferID);
}

} // namespace rendering
} // namespace openGL
