#include "curveapp.h"

#include <algorithm>
#include <cassert>
#include <iostream>
#include <limits>
#include <numeric>

#include "curvefileio.h"
#include "openglmatrix.h"
#include "opengltools.h"
#include "plane.h"
#include "shaderSource.h"
#include "vec4f.h"

using namespace openGL;

namespace animation {

CurveApp::CurveApp(int argc, char const *argv[]) {
  if (argc == 2) {
    m_loadFilePath = std::string(argv[1]);
    m_saveFilePath = m_loadFilePath;
  }
  if (argc == 3) {
    m_loadFilePath = std::string(argv[1]);
    m_saveFilePath = std::string(argv[2]);
  }
}

CurveApp::~CurveApp() {
  // clear resources before destroying context
  m_resourceManager.clear();

  glfwDestroyWindow(m_window);
  m_window = nullptr; // invalidate
  glfwTerminate();
}

bool CurveApp::initialize() {
  if (!setupWindow())
    return false;

  if (!setupShaders())
    return false;

  if (!setupCurveBuffer())
    return false;

  if (!setupAxisBuffer())
    return false;

  setupCallbackFunctions();

  glClearColor(0.25f, 0.25f, 0.25f, 1.f);

  glEnable(GL_MULTISAMPLE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_POINT_SPRITE);
  glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
  glDisable(GL_CULL_FACE);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  loadCurveFromFile(m_loadFilePath);
  loadCurveBuffer();

  loadAxisBuffer();

  resetCamera();
  reloadModelMatrix();
  reloadViewMatrix();
  reloadProjectionMatrix();
  reloadMVPMatrix();
  reloadMVPUniform();

  return true;
}

bool CurveApp::run() {
  if (!m_window)
    return false;

  while (!glfwWindowShouldClose(m_window)) {
    oncePerFrame();
    glfwSwapBuffers(m_window);
    glfwPollEvents();
  }

  return true;
}

bool CurveApp::setupWindow() {
  if (!glfwInit()) {
    std::cerr << "Failed to initialize GLFW\n";
    return false;
  }

  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
  glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_FALSE);
  glfwWindowHint(GLFW_SAMPLES, 4);

  m_window = glfwCreateWindow(m_viewport.width(), m_viewport.height(),
                              "CurveApp: CPSC587/687", nullptr, nullptr);

  if (!m_window) {
    std::cerr << "Failed to create OpenGL context\n";
    glfwTerminate();
    return false;
  }

  // just incase framebuffer size differs (e.g. retina display)
  math::Vec2i tmpSize;
  glfwGetFramebufferSize(m_window, &tmpSize.m_x, &tmpSize.m_y);
  m_viewport.framebufferSize(tmpSize);

  glfwMakeContextCurrent(m_window);
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);
  glfwSwapInterval(1);

  return true;
}

bool CurveApp::setupShaders() {
  using namespace rendering;

  // make point shader (1)
  {
    auto vs = makeShader(glslSource::PointVS_Source, Shader::VERTEX);
    if (!vs) {
      std::cerr << "Failed to compile Point (vertex) shader\n";
      return false;
    }
    auto fs = makeShader(glslSource::PointFS_Source, Shader::FRAGMENT);
    if (!fs) {
      std::cerr << "Failed to compile Point (fragment) shader\n";
      return false;
    }
    auto pr = makeProgram(*vs, *fs);
    if (!pr) {
      std::cerr << "Failed to link Point shader\n";
      return false;
    }
    int pointID = m_resourceManager.pushProgram(std::move(*pr));
    assert(pointID == ShaderID::POINT);

    // setup inital uniforms
    auto &program = m_resourceManager.getProgram(ShaderID::POINT);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
    program.setUniform1f("pointSize", 20.f);
  }

  // make line shader (2)
  {
    auto vs = makeShader(glslSource::LineVS_Source, Shader::VERTEX);
    if (!vs) {
      std::cerr << "Failed to compile Point (vertex) shader\n";
      return false;
    }
    auto gs = makeShader(glslSource::LineGS_Source, Shader::GEOMETRY);
    if (!gs) {
      std::cerr << "Failed to compile Point (geometry) shader\n";
      return false;
    }
    auto fs = makeShader(glslSource::LineFS_Source, Shader::FRAGMENT);
    if (!fs) {
      std::cerr << "Failed to compile Point (fragment) shader\n";
      return false;
    }
    auto pr = makeProgram(*vs, *gs, *fs);
    if (!pr) {
      std::cerr << "Failed to link Point shader\n";
      return false;
    }
    int lineID = m_resourceManager.pushProgram(std::move(*pr));
    assert(lineID == ShaderID::LINE);

    // setup inital uniforms
    auto &program = m_resourceManager.getProgram(ShaderID::LINE);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
    program.setUniform2i("windowSize", m_viewport.windowSize());
    program.setUniform1f("lineThickness", 10.f);
  }

  // make axis shader (3)
  {
    auto vs = makeShader(glslSource::AxisVS_Source, Shader::VERTEX);
    if (!vs) {
      std::cerr << "Failed to compile Axis (vertex) shader\n";
      return false;
    }
    auto fs = makeShader(glslSource::AxisFS_Source, Shader::FRAGMENT);
    if (!fs) {
      std::cerr << "Failed to compile Axis (fragment) shader\n";
      return false;
    }
    auto pr = makeProgram(*vs, *fs);
    if (!pr) {
      std::cerr << "Failed to link Axis shader\n";
      return false;
    }
    int pointID = m_resourceManager.pushProgram(std::move(*pr));
    assert(pointID == ShaderID::AXIS);

    // setup inital uniforms
    auto &program = m_resourceManager.getProgram(ShaderID::AXIS);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
  }

  return true;
}

bool CurveApp::setupCurveBuffer() {
  using namespace rendering;

  auto cb = makeCurveBuffer();
  if (!cb) {
    std::cerr << "Failed to allocate Curve buffer\n";
    return false;
  }
  int curveID = m_resourceManager.pushCurveBuffer(std::move(*cb));
  assert(curveID == CurveID::BASE);

  return true;
}

bool CurveApp::setupAxisBuffer() {
  using namespace rendering;

  auto ab = makeAxisBuffer();
  if (!ab) {
    std::cerr << "Failed to allocate Axis buffer\n";
    return false;
  }
  int axisID = m_resourceManager.pushAxisBuffer(std::move(*ab));
  assert(axisID == AxisID::BASE);

  return true;
}

void CurveApp::oncePerFrame() { render(); }

void CurveApp::render() {
  glBindFramebuffer(GL_FRAMEBUFFER, 0); // default framebuffer

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  auto &pointProgram = m_resourceManager.getProgram(ShaderID::POINT);
  pointProgram.use();

  auto &curveBuffer = m_resourceManager.getCurveBuffer(CurveID::BASE);
  curveBuffer.bindVAO();
  glDrawArrays(GL_POINTS, 0, curveBuffer.size());

  auto &lineProgram = m_resourceManager.getProgram(ShaderID::LINE);
  lineProgram.use();
  if (m_curve.isClosed())
    glDrawArrays(GL_LINE_LOOP, 0, curveBuffer.size());
  else
    glDrawArrays(GL_LINE_STRIP, 0, curveBuffer.size());

  auto &axisBuffer = m_resourceManager.getAxisBuffer(AxisID::BASE);
  axisBuffer.bindVAO();
  auto &axisProgram = m_resourceManager.getProgram(ShaderID::AXIS);
  axisProgram.use();
  glDrawArrays(GL_LINES, 0, axisBuffer.size());
}

void CurveApp::resetCamera() { m_camera = scene::Camera{{0.f, 0.f, 1.f}}; }

void CurveApp::loadCurveFromFile(std::string const &filePath) {
  auto isCurve = math::geometry::loadCurveFromFile(filePath);
  if (!isCurve) {
    // default
    m_curve = math::geometry::Curve({{0.5f, 0.f, 0.5f},
                                     {0.5f, 0.f, -0.5f},
                                     {-0.5f, 0.f, -0.5f},
                                     {-0.5f, 0.f, 0.5f}});
    m_curve.setClosed(true);
  } else {
    m_curve = (*isCurve);
    m_curve.setClosed(true);
  }

  m_selectedID.clear();
}

void CurveApp::saveCurveToFile(std::string const &filePath) {
  math::geometry::saveCurveToFile(m_curve, filePath);
}

void CurveApp::loadCurveBuffer() {
  auto &curveBuffer = m_resourceManager.getCurveBuffer(CurveID::BASE);
  curveBuffer.setupCurveBuffers(m_curve.points(), {1.f, 0.f, 0.f});
}

void CurveApp::loadSelectedBuffer() {
  auto &curveBuffer = m_resourceManager.getCurveBuffer(CurveID::BASE);

  std::vector<math::Vec3f> colors;
  colors.resize(curveBuffer.size(), {1.f, 0.f, 0.f});
  for (auto idx : m_selectedID) {
    colors[idx] = {0.f, 1.f, 1.f};
  }

  if (!m_selectedID.empty())
    colors[m_selectedID.back()] = {0.3f, 0.4f, 1.f};

  curveBuffer.loadColorBuffer(colors);
}

void CurveApp::loadAxisBuffer() {
  auto &axisBuffer = m_resourceManager.getAxisBuffer(AxisID::BASE);
  axisBuffer.setupAxisBuffer(m_selectedOrigins);
}

void CurveApp::reloadModelMatrix() {
  using namespace math;
  m_M = math::identity();
}

void CurveApp::reloadViewMatrix() { m_V = makeViewMatrix(m_camera); }

void CurveApp::reloadProjectionMatrix() {
  m_P = makeProjectionMatrix(m_viewport);
}

void CurveApp::reloadMVPMatrix() { m_MVP = m_P * m_V * m_M; }

void CurveApp::reloadMVPUniform() {
  // TODO use uniform blocks to share VP between shaders
  {
    auto &program = m_resourceManager.getProgram(ShaderID::POINT);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
  }

  {
    auto &program = m_resourceManager.getProgram(ShaderID::LINE);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
  }

  {
    auto &program = m_resourceManager.getProgram(ShaderID::AXIS);
    program.use();
    program.setUniformMat4f("MVP", m_MVP, true);
  }
}

void CurveApp::selectPointAround(math::Vec2i const &pixel) {
  using namespace std;

  auto id = findClosestProjectPointTo(pixel, m_curve.points(), 10.f);
  if (id != -1) {
    std::cerr << "point id: " << id << " at position: " << m_curve[id] << '\n';
    // Linear search is fine (quickest) for small vectors
    auto insertPos = find(begin(m_selectedID), end(m_selectedID), id);
    if (insertPos == end(m_selectedID)) {
      m_selectedID.push_back(id);
    } else {
      m_selectedID.erase(insertPos);
    }
  }
}

void CurveApp::moveSelectedTo(const math::Vec2i &pixel) {
  if (!m_selectedID.empty()) {
    auto &pos = m_curve[m_selectedID.back()];
    auto ray = rayFromPixelPosition(pixel, m_camera, m_viewport);
    auto interesction = math::geometry::intersectRayPlane(
        pos, m_camera.localForward(), ray.m_origin, ray.m_direction, 1e-8);
    if (interesction) {
      math::Vec3f translation = (*interesction).position - pos;
      applyTranslationToSelected(translation);
    }
  }
}

void CurveApp::deleteSelected() {
  std::vector<int> tmp(m_curve.pointCount());
  std::vector<math::Vec3f> tmpCurvePoints;
  tmpCurvePoints.reserve(m_curve.pointCount() - m_selectedID.size());

  std::iota(begin(tmp), end(tmp), 0);

  for (auto idx : m_selectedID) {
    tmp[idx] = -1; // set invalid
  }

  tmp.erase(std::stable_partition(begin(tmp), end(tmp),
                                  [](int idx) { return idx != -1; }),
            end(tmp));
  for (auto idx : tmp) {
    tmpCurvePoints.emplace_back(m_curve[idx]);
  }

  m_curve = {tmpCurvePoints};
  m_curve.setClosed(true);

  m_selectedID.clear();
}

void CurveApp::subdivideSelected() {
  // Put in ascending order
  std::sort(begin(m_selectedID), end(m_selectedID));
  // subdividing in the reverse order does not invalidate IDs
  std::reverse(begin(m_selectedID), end(m_selectedID));

  for (auto idx : m_selectedID) {
    m_curve.addMidpointToSegment(idx);
  }

  m_selectedID.clear();
}

void CurveApp::applyTranslationToSelected(const math::Vec3f &offset) {
  for (auto idx : m_selectedID) {
    m_curve[idx] += offset;
  }
}

int CurveApp::findClosestProjectPointTo(math::Vec2i const &pixel,
                                        const std::vector<math::Vec3f> &points,
                                        float pixelThreshold) {
  using namespace math;

  float const pixelThresholdSquared = pixelThreshold * pixelThreshold;

  int currentClosetId = -1; // sentinel value
  float currentMinimumDistanceSqrd = pixelThresholdSquared;

  /**
   * For derivation and pictures see
   * http://antongerdelan.net/opengl/raycasting.html
   */

  for (size_t id = 0; id < points.size(); ++id) {
    Vec4f homogeneousClipSpace = m_MVP * Vec4f(points[id]);
    Vec3f normalizedDeviceCoordinates = perspectiveDivide(homogeneousClipSpace);
    Vec2i projectedPoint =
        pixelPosition(normalizedDeviceCoordinates, m_viewport);

    float distanceSqrd = distanceSquared(projectedPoint, pixel);

    if (distanceSqrd <= currentMinimumDistanceSqrd) {
      currentClosetId = id;
      currentMinimumDistanceSqrd = distanceSqrd;
    }
  }

  return currentClosetId;
}

math::Vec3f CurveApp::calculateScreenPoint(math::Vec2i const &pixel,
                                           math::Vec3f const &origin) {
  auto ray = rayFromPixelPosition(pixel, m_camera, m_viewport);
  auto interesction = math::geometry::intersectRayPlane(
      origin, m_camera.localForward(), ray.m_origin, ray.m_direction, 1e-8);
  if (interesction) {
    return (*interesction).position;
  }
  return origin;
}

void CurveApp::recalculateSelectedAxisOrigin() {
  if (m_selectedID.empty()) {
    m_selectedOrigins[0] = {0.f, 0.f, 0.f};
  } else {
    math::Vec3f median;

    for (auto idx : m_selectedID) {
      median += m_curve[idx];
    }
    median /= m_selectedID.size();

    m_selectedOrigins[0] = median;
  }
}

int CurveApp::removePoint(math::geometry::Curve &curve, int idx) {
  curve.removePoint(idx);
  return 0;
}

int CurveApp::addPoint(math::geometry::Curve &curve, int idx) {
  curve.addMidpointToSegment(idx);
  return 0;
}

// Callback functions
void CurveApp::setupCallbackFunctions() {
  glfwSetWindowUserPointer(m_window, this);

  glfwSetWindowSizeCallback(m_window, [](GLFWwindow *w, int x, int y) {
    static_cast<CurveApp *>(glfwGetWindowUserPointer(w))->onSetWindowSize(x, y);
  });

  glfwSetFramebufferSizeCallback(m_window, [](GLFWwindow *w, int x, int y) {
    static_cast<CurveApp *>(glfwGetWindowUserPointer(w))
        ->onSetFramebufferSize(x, y);
  });

  glfwSetKeyCallback(
      m_window, [](GLFWwindow *w, int key, int scancode, int action, int mods) {
        static_cast<CurveApp *>(glfwGetWindowUserPointer(w))
            ->onKeyPress(key, scancode, action, mods);
      });

  glfwSetMouseButtonCallback(
      m_window, [](GLFWwindow *w, int button, int action, int mods) {
        static_cast<CurveApp *>(glfwGetWindowUserPointer(w))
            ->onMouseButton(button, action, mods);
      });

  glfwSetCursorPosCallback(m_window, [](GLFWwindow *w, double x, double y) {
    static_cast<CurveApp *>(glfwGetWindowUserPointer(w))->onMouseMotion(x, y);
  });
}

void CurveApp::onSetWindowSize(int width, int height) {

  m_viewport.windowSize({width, height});

  auto &program = m_resourceManager.getProgram(ShaderID::LINE);
  program.use();
  program.setUniform2i("windowSize", m_viewport.windowSize());

  reloadProjectionMatrix();
  reloadMVPMatrix();
  reloadMVPUniform();

  glViewport(0, 0, m_viewport.width(), m_viewport.height());
}

void CurveApp::onSetFramebufferSize(int width, int height) {
  m_viewport.framebufferSize({width, height});
}

void CurveApp::onMouseButton(int button, int action, int mods) {
  if (mods == GLFW_MOD_CONTROL && action != GLFW_RELEASE) {
    m_mode = InputMode::CAMERA_TURNTABLE;
  } else if (button == GLFW_MOUSE_BUTTON_MIDDLE && action != GLFW_RELEASE) {
    m_mode = InputMode::CAMERA_ZOOM;
  } else {
    double x, y;
    glfwGetCursorPos(m_window, &x, &y);
    if (action == GLFW_PRESS && mods == GLFW_MOD_SHIFT) {
      m_mode = InputMode::EDIT;
      selectPointAround({x, y});
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    } else if (action == GLFW_RELEASE) {
      m_mode = InputMode::NONE;
      m_translationOffset = {0.f, 0.f, 0.f};
    } else {
      m_mode = InputMode::EDIT;
      if (m_selectedID.size() == 1)
        m_selectedID.clear();
      selectPointAround({x, y});
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
      if (!m_selectedID.empty()) {
        m_translationOffset =
            calculateScreenPoint({x, y}, m_curve[m_selectedID.back()]);
        m_translationOffset -= m_curve[m_selectedID.back()];
      }
    }
  }
}

void CurveApp::onMouseMotion(double x, double y) {
  static math::Vec2i previous{x, y};
  if (m_mode == InputMode::CAMERA_TURNTABLE) {
    float deltaX = (x - previous.m_x) / m_viewport.width();
    float deltaY = (y - previous.m_y) / m_viewport.height();

    constexpr float speed = 50.f;
    m_camera.rotateLeftAroundFocus(deltaX * speed);
    m_camera.rotateUpAroundFocus(deltaY * speed);

    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } else if (m_mode == InputMode::CAMERA_ZOOM) {
    float deltaY = (y - previous.m_y) / m_viewport.height();

    constexpr float speed = 5.f;
    m_camera.moveTowardFocus(deltaY * speed);

    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } else {
    if (m_mode == InputMode::EDIT) {
      moveSelectedTo({x, y});
      applyTranslationToSelected(-m_translationOffset);
      loadCurveBuffer();
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    }
  }

  previous = {x, y};
}

void CurveApp::onKeyPress(int key, int scancode, int action, int mods) {

  constexpr float cameraMoveSpeed = 0.1f;
  constexpr float cameraRotateSpeed = 10.f;

  if (action == GLFW_RELEASE)
    return;

  switch (key) {
  case GLFW_KEY_ESCAPE: {
    glfwSetWindowShouldClose(m_window, GL_TRUE);
    break;
  }
  case GLFW_KEY_P: {
    m_curve = math::geometry::repeatedAveraging(m_curve.points(), 3);
    m_curve.setClosed(true);

    loadCurveBuffer();
    break;
  }
  case GLFW_KEY_L: {
    std::vector<math::Vec3f> tmpCurve = {
        {0.5f, 0.f, 0.5f},
        {0.5f, 0.f, -0.5f},
        {-0.5f, 0.f, -0.5f},
        {-0.5f, 0.f, 0.5f},
    };
    m_curve = tmpCurve;
    m_curve.setClosed(true);
    loadCurveBuffer();
  } break;
  case GLFW_KEY_M: {
    if (mods == GLFW_MOD_CONTROL) {
      deleteSelected();
      loadCurveBuffer();
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    }
  } break;
  case GLFW_KEY_C: {
    if (mods == GLFW_MOD_CONTROL) {
      subdivideSelected();
      loadCurveBuffer();
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    }
  } break;
  case GLFW_KEY_W: {
    m_camera.moveForward(cameraMoveSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_O: {
    if (mods == GLFW_MOD_CONTROL) {
      loadCurveFromFile(m_loadFilePath);
      loadCurveBuffer();
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    }
  } break;
  case GLFW_KEY_S: {
    if (mods == GLFW_MOD_CONTROL) {
      saveCurveToFile(m_saveFilePath);
    } else {
      m_camera.moveBackward(cameraMoveSpeed);
      reloadViewMatrix();
      reloadMVPMatrix();
      reloadMVPUniform();
    }
  } break;
  case GLFW_KEY_A: {
    m_camera.moveLeft(cameraMoveSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_D: {
    if (mods == GLFW_MOD_CONTROL) {
      m_selectedID.clear();
      loadSelectedBuffer();
      recalculateSelectedAxisOrigin();
      loadAxisBuffer();
    } else {
      m_camera.moveRight(cameraMoveSpeed);
      reloadViewMatrix();
      reloadMVPMatrix();
      reloadMVPUniform();
    }
  } break;
  case GLFW_KEY_E: {
    m_camera.moveUp(cameraMoveSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_Q: {
    m_camera.moveDown(cameraMoveSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_LEFT: {
    if (mods == GLFW_MOD_SHIFT)
      m_camera.rollLeft(cameraRotateSpeed);
    else
      m_camera.rotateLeft(cameraRotateSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_RIGHT: {
    if (mods == GLFW_MOD_SHIFT)
      m_camera.rollRight(cameraRotateSpeed);
    else
      m_camera.rotateRight(cameraRotateSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_UP: {
    m_camera.rotateUp(cameraRotateSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_DOWN: {
    m_camera.rotateDown(cameraRotateSpeed);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_1: {
    float dist = distance(m_camera.localPos(), {0.f, 0.f, 0.f});
    if (mods == GLFW_MOD_SHIFT)
      m_camera =
          scene::Camera({dist, 0.f, 0.f}, {-1.f, 0.f, 0.f}, {0.f, 1.f, 0.f});
    else
      m_camera =
          scene::Camera({-dist, 0.f, 0.f}, {1.f, 0.f, 0.f}, {0.f, 1.f, 0.f});

    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_2: {
    float dist = distance(m_camera.localPos(), {0.f, 0.f, 0.f});
    if (mods == GLFW_MOD_SHIFT)
      m_camera =
          scene::Camera({0.f, -dist, 0.f}, {0.f, 1.f, 0.f}, {0.f, 0.f, 1.f});
    else
      m_camera =
          scene::Camera({0.f, dist, 0.f}, {0.f, -1.f, 0.f}, {0.f, 0.f, -1.f});
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_3: {
    float dist = distance(m_camera.localPos(), {0.f, 0.f, 0.f});
    if (mods == GLFW_MOD_SHIFT)
      m_camera =
          scene::Camera({0.f, 0.f, dist}, {0.f, 0.f, -1.f}, {0.f, 1.f, 0.f});
    else
      m_camera =
          scene::Camera({0.f, 0.f, -dist}, {0.f, 0.f, 1.f}, {0.f, 1.f, 0.f});
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;
  case GLFW_KEY_4: {
    m_camera.setFocusPoint(m_selectedOrigins[0]);
    reloadViewMatrix();
    reloadMVPMatrix();
    reloadMVPUniform();
  } break;

  default:
    break;
  }
}

} // namespace animation
