#include "curvebuffer.h"

#include <cassert>

namespace openGL {
namespace rendering {

CurveBuffer::~CurveBuffer() { release(); }

CurveBuffer::CurveBuffer(GLuint vaoID, GLuint vertexBufferID,
                         GLuint colorBufferID)
    : m_vaoID(vaoID), m_vertexBufferID(vertexBufferID),
      m_colorBufferID(colorBufferID) {
  setupVAO();
}

CurveBuffer::CurveBuffer(CurveBuffer &&other)
    : m_vaoID(other.m_vaoID), m_vertexBufferID(other.m_vertexBufferID),
      m_colorBufferID(other.m_colorBufferID), m_bufferSize(other.m_bufferSize) {
  other.m_vaoID = 0;
  other.m_vertexBufferID = 0;
  other.m_colorBufferID = 0;
  other.m_bufferSize = 0;
}

CurveBuffer &CurveBuffer::operator=(CurveBuffer &&other) {
  if (this != &other) {
    release();
    std::swap(m_vaoID, other.m_vaoID);
    std::swap(m_vertexBufferID, other.m_vertexBufferID);
    std::swap(m_colorBufferID, other.m_colorBufferID);
    std::swap(m_bufferSize, other.m_bufferSize);
  }
  return *this;
}

void CurveBuffer::bindVAO() { glBindVertexArray(m_vaoID); }

void CurveBuffer::setupVAO() {
  bindVAO();

  // position attribute
  glEnableVertexAttribArray(0); // layout=0
  bindVertex();
  glVertexAttribPointer(0,        // layout=0
                        3,        // # of components
                        GL_FLOAT, // type of components
                        GL_FALSE, // normalize
                        0,        // stride
                        (void *)0 // array buffer offset
  );

  // Color attribute
  glEnableVertexAttribArray(1); // layout=1
  bindColor();
  glVertexAttribPointer(1,        // layout=1
                        3,        // # of components
                        GL_FLOAT, // type of components
                        GL_FALSE, // normalize
                        0,        // stride
                        (void *)0 // array buffer offset
  );

  // reset to default
  glBindVertexArray(0);
}

void CurveBuffer::setupCurveBuffers(std::vector<math::Vec3f> const &vertices,
                                    std::vector<math::Vec3f> const &colors) {
  assert(vertices.size() == colors.size());

  bindVertex();
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * vertices.size(),
               vertices.data(), GL_DYNAMIC_DRAW);

  bindColor();
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * colors.size(),
               colors.data(), GL_DYNAMIC_DRAW);

  m_bufferSize = vertices.size();
}

void CurveBuffer::setupCurveBuffers(std::vector<math::Vec3f> const &vertices,
                                    math::Vec3f const &defaultColorRGB) {
  bindVertex();
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * vertices.size(),
               vertices.data(), GL_DYNAMIC_DRAW);

  bindColor();
  std::vector<math::Vec3f> colors(vertices.size(), defaultColorRGB);
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * colors.size(),
               colors.data(), GL_DYNAMIC_DRAW);

  m_bufferSize = vertices.size();
}

void CurveBuffer::loadColorBuffer(std::vector<math::Vec3f> const &colors) {
  assert(m_bufferSize == colors.size());

  bindColor();
  glBufferData(GL_ARRAY_BUFFER, sizeof(math::Vec3f) * colors.size(),
               colors.data(), GL_DYNAMIC_DRAW);
}

size_t CurveBuffer::size() const { return m_bufferSize; }

void CurveBuffer::release() {
  glDeleteVertexArrays(1, &m_vaoID);
  glDeleteBuffers(1, &m_vertexBufferID);
  glDeleteBuffers(1, &m_colorBufferID);

  m_vaoID = 0;
  m_vertexBufferID = 0;
  m_colorBufferID = 0;
}

void CurveBuffer::bindVertex() {
  glBindBuffer(GL_ARRAY_BUFFER, m_vertexBufferID);
}

void CurveBuffer::bindColor() {
  glBindBuffer(GL_ARRAY_BUFFER, m_colorBufferID);
}

std::optional<CurveBuffer> makeCurveBuffer() {
  GLuint vaoID = 0;
  GLuint vertexBufferID = 0;
  GLuint colorBufferID = 0;
  glGenVertexArrays(1, &vaoID);
  glGenBuffers(1, &vertexBufferID);
  glGenBuffers(1, &colorBufferID);

  if (vaoID == 0 || vertexBufferID == 0 || colorBufferID == 0)
    return std::nullopt;

  return CurveBuffer(vaoID, vertexBufferID, colorBufferID);
}

} // namespace rendering
} // namespace openGL
