#include "vec4f.h"

#include <cmath>
#include <iostream>

namespace math {

// Member function interface
Vec4f::Vec4f(float x, float y, float z, float w)
    : m_x(x), m_y(y), m_z(z), m_w(w) {}

// TODO: gcc bug
Vec4f::Vec4f(Vec3f const &v)
    : m_x(v.m_x), m_y(v.m_y), m_z(v.m_z),
      m_w(1.f) // should not have to initialize, but defaults to 0 otherwise
{}

Vec4f::Vec4f(Vec3f const &v, float w)
    : m_x(v.m_x), m_y(v.m_y), m_z(v.m_z), m_w(w) {}

Vec4f::operator Vec3f() const { return {m_x, m_y, m_z}; }

Vec4f &Vec4f::operator+=(Vec4f const &rhs) {
  m_x += rhs.m_x;
  m_y += rhs.m_y;
  m_z += rhs.m_z;
  m_w += rhs.m_w;
  return *this;
}

Vec4f &Vec4f::operator-=(Vec4f const &rhs) {
  m_x -= rhs.m_x;
  m_y -= rhs.m_y;
  m_z -= rhs.m_z;
  m_w -= rhs.m_w;
  return *this;
}

Vec4f &Vec4f::operator*=(float rhs) {
  m_x *= rhs;
  m_y *= rhs;
  m_z *= rhs;
  m_w *= rhs;
  return *this;
}

Vec4f &Vec4f::operator/=(float rhs) {
  m_x /= rhs;
  m_y /= rhs;
  m_z /= rhs;
  m_w /= rhs;
  return *this;
}

float &Vec4f::operator[](size_t id) { return m_coord[id]; }

float const &Vec4f::operator[](size_t id) const { return m_coord[id]; }

float *Vec4f::data() { return &m_coord[0]; }

float const *Vec4f::data() const { return &m_coord[0]; }

Vec4f &Vec4f::normalize() {
  float l = norm(*this);
  return (*this) /= l;
}

void Vec4f::zero() {
  m_x = 0.f;
  m_y = 0.f;
  m_z = 0.f;
  m_w = 0.f;
}

// Free function
/*
 * Vector-Vector Addition/ Subtraction
 */
Vec4f operator+(Vec4f const &a, Vec4f const &b) {
  return Vec4f(a.m_x + b.m_x, a.m_y + b.m_y, a.m_z + b.m_z, a.m_w + b.m_w);
}

Vec4f operator-(Vec4f const &a, Vec4f const &b) {
  return Vec4f(a.m_x - b.m_x, a.m_y - b.m_y, a.m_z - b.m_z, a.m_w - b.m_w);
}

/*
 * Scalar-Vector Multiplication/Division
 */
Vec4f operator*(float s, Vec4f v) {
  v.m_x *= s;
  v.m_y *= s;
  v.m_z *= s;
  v.m_w *= s;
  return v;
}
Vec4f operator*(Vec4f v, float s) { return s * v; }

Vec4f operator/(Vec4f v, float s) {
  v.m_x /= s;
  v.m_y /= s;
  v.m_z /= s;
  v.m_w /= s;
  return v;
}

/*
 * Negation of vector
 * -v = (-1.f) * v
 */
Vec4f operator-(Vec4f v) {
  v.m_x = -v.m_x;
  v.m_y = -v.m_y;
  v.m_z = -v.m_z;
  v.m_w = -v.m_w;
  return v;
}

/*
 * Vector-Vector (inner/dot) product
 */
float operator*(Vec4f const &a, Vec4f const &b) {
  return a.m_x * b.m_x + a.m_y * b.m_y + a.m_z * b.m_z + a.m_w * b.m_w;
}

float dot(Vec4f const &a, Vec4f const &b) { return a * b; }

/*
 * Vector norm (length)
 */
float norm(Vec4f const &v) {
  return std::sqrt(v.m_x * v.m_x + v.m_y * v.m_y + v.m_z * v.m_z +
                   v.m_w * v.m_w);
}

float normSquared(Vec4f const &v) {
  return v.m_x * v.m_x + v.m_y * v.m_y + v.m_z * v.m_z + v.m_w * v.m_w;
}

/*
 * Normalized Vector
 */
Vec4f normalized(Vec4f v) {
  float l = norm(v);
  return v /= l;
}

Vec4f componentMultiplication(const Vec4f &lhs, const Vec4f &rhs) {
  return {lhs.m_x * rhs.m_x, lhs.m_y * rhs.m_y, lhs.m_z * rhs.m_z,
          lhs.m_w * rhs.m_w};
}

Vec4f operator*(Mat4f const &lhs, Vec4f const &rhs) {
  Vec4f tmp;
  float element;

  for (int r = 0; r < Mat4f::DIM; ++r) {
    element = 0.f;
    for (int c = 0; c < Mat4f::DIM; ++c) {
      float l = lhs(r, c);
      float r = rhs[c];
      element += l * r;
    }
    tmp[r] = element;
  }

  return tmp;
}

std::ostream &operator<<(std::ostream &out, Vec4f const &v) {
  return out << v.m_x << " " << v.m_y << " " << v.m_z << " " << v.m_w;
}

std::istream &operator>>(std::istream &in, Vec4f &v) {
  return in >> v.m_x >> v.m_y >> v.m_z >> v.m_w;
}

} // namespace math
