#include "viewport.h"
#include "openglmatrix.h"

#include <cmath>

using namespace math;

namespace openGL {
namespace rendering {

Viewport::Viewport(Vec2i size) : m_windowSize(size), m_framebufferSize(size) {}

Viewport::Viewport(Vec2i size, float fovDegrees, float nearZ, float farZ)
    : m_windowSize(size), m_framebufferSize(size), m_fovDegrees(fovDegrees),
      m_nearZ(nearZ), m_farZ(farZ) {}

void Viewport::windowSize(Vec2i size) { m_windowSize = size; }

void Viewport::framebufferSize(Vec2i size) { m_framebufferSize = size; }

void Viewport::nearPlaneDistance(float distance) { m_nearZ = distance; }

void Viewport::farPlaneDistance(float distance) { m_farZ = distance; }

void Viewport::fieldOfViewDegrees(float degrees) { m_fovDegrees = degrees; }

Vec2i Viewport::windowSize() const { return m_windowSize; }

Vec2i Viewport::framebufferSize() const { return m_framebufferSize; }

int Viewport::width() const { return m_windowSize.m_x; }

int Viewport::height() const { return m_windowSize.m_y; }

float Viewport::aspectRatio() const {
  return static_cast<float>(m_windowSize.m_x) / m_windowSize.m_y;
}

float Viewport::nearPlaneDistance() const { return m_nearZ; }

float Viewport::farPlaneDistance() const { return m_farZ; }

float Viewport::fieldOfViewDegrees() const { return m_fovDegrees; }

// Free functions

Mat4f makeProjectionMatrix(Viewport const &viewport) {
  return PerspectiveProjection(
      viewport.fieldOfViewDegrees(), viewport.aspectRatio(),
      viewport.nearPlaneDistance(), viewport.farPlaneDistance());
}

Mat4f makeInverseProjectionMatrix(Viewport const &viewport) {
  return inversePerspectiveProjection(
      viewport.fieldOfViewDegrees(), viewport.aspectRatio(),
      viewport.nearPlaneDistance(), viewport.farPlaneDistance());
}

Vec3f normalizedDeviceCoordinates(Vec2i const &pixelPosition,
                                  Viewport const &viewport) {
  return {(2.f * pixelPosition.m_x) / viewport.width() - 1.f,
          (-2.f * pixelPosition.m_y) / viewport.height() + 1.f, 0.f};
}

Vec2i pixelPosition(Vec3f const &normalizedDeviceCoordinates,
                    Viewport const &viewport) {
  return {viewport.width() * 0.5f * (normalizedDeviceCoordinates.m_x + 1),
          viewport.height() * 0.5f * (1 - normalizedDeviceCoordinates.m_y)};
}

Vec3f worldDimensions(Viewport const &viewport) {
  float width = std::tan(viewport.fieldOfViewDegrees() * (M_PI / 180.f)) *
                viewport.nearPlaneDistance();
  float height = width / viewport.aspectRatio();
  return {width, height, 0.f};
}

} // namespace rendering
} // namespace openGL
